#  Copyright 2023 Fabarta Authors.
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import json
from pyspark.sql.types import StructField, LongType, StringType

from config import Config

SCHEMA_DIR = 'schema'
VERTEX_SCHEMA = 'vertexes.schema'
EDGE_SCHEMA = 'edges.schema'
OID_COLUMN_NAME = '_oid'
EDGE_SRC_ID_COLUMN = 'src_id'
EDGE_DST_ID_COLUMN = 'dst_id'
EDGE_RANK_COLUMN = 'rank'

def read_schema(config: Config):
    with open(os.path.join(config.tmp_dir, SCHEMA_DIR, VERTEX_SCHEMA), 'r') as vf:
        vertex_schema = json.load(vf)

    with open(os.path.join(config.tmp_dir, SCHEMA_DIR, EDGE_SCHEMA), 'r') as ef:
        edge_schema = json.load(ef)

    return vertex_schema, edge_schema


def get_edge_full_names(edge_schema: dict):
    edge_full_names = []
    for pair in edge_schema['vertex_type_pair']:
        edge_full_name = '{}_{}_{}'.format(pair['from_name'],
                                           edge_schema['edge_type_name'], pair['to_name'])
        edge_full_names.append(edge_full_name)
    return edge_full_names


def build_vertex_fields(schema, columns):
    fields = []
    col_idx = 0
    for prop in schema['properties']:
        column_name = prop['property_name']
        if column_name == OID_COLUMN_NAME:
            continue

        if len(columns) > 0:
            column_name = columns[col_idx]
            col_idx += 1

        if prop['primary_key']:
            if prop['property_type'] == 'int64':
                fields.append(StructField(column_name, LongType(), False))
            else:
                fields.append(StructField(column_name, StringType(), False))
        else:
            fields.append(StructField(column_name, StringType(), prop['nullable']))
    return fields


def build_edge_fields(edge_schema, columns, vertex_schemas, src_vertex_name, dst_vertex_name):
    fields = []
    src_vertex = json.loads(vertex_schemas[src_vertex_name])
    fields.append(StructField(EDGE_SRC_ID_COLUMN, LongType() if src_vertex['key_type'] == 'int64' \
        else StringType(), False))
    dst_vertex = json.loads(vertex_schemas[dst_vertex_name])
    fields.append(StructField(EDGE_DST_ID_COLUMN, LongType() if dst_vertex['key_type'] == 'int64' \
        else StringType(), False))

    # skip first two columns
    col_idx = 2
    for prop in edge_schema['properties']:
        if len(columns) > 0:
            column_name = columns[col_idx]
            col_idx += 1
        else:
            column_name = prop['property_name']
        fields.append(StructField(column_name, StringType(), prop['nullable']))
    # the rank column MUST be the last column
    if edge_schema['rankable']:
        fields.append(StructField(EDGE_RANK_COLUMN, StringType(), False))

    return fields

''' 
<<<<<<<< vertex schema example >>>>>>>>
"person":"{
    "graph_id":1025,
    "vertex_type_id":1,
    "vertex_type_name":"person",
    "schema_version":0,
    "data_version":0,
    "status":"Normal",
    "key_type":"int64",
    "key_length":8,
    "properties":[{
            "property_id":1,
            "property_name":"_oid",
            "property_type":"uint64",
            "type_length":8,
            "nullable":false,
            "primary_key":false,
            "default_value":{"object":null},
            "index_num":0,
            "order":0,
            "comment":"",
            "is_unique":true,
            "is_full_text":false,
            "v_index_type":null,
            "is_oid":true
        },{
            "property_id":2,
            "property_name":"id",
            "property_type":"int64",
            "type_length":8,
            "nullable":false,
            "primary_key":true,
            "default_value":{"object":null},
            "index_num":1,
            "order":0,
            "comment":"",
            "is_unique":true,
            "is_full_text":false,
            "v_index_type":null,
            "is_oid":false
        }],
    "partition_schema":{
        "method":"hash",
        "column":"_TYPE"
    },
    "rankable":false,
    "comment":"person",
    "indexes":[{
        "index_id":2,
        "index_name":"person_pk_index",
        "index_type":"Vertex",
        "graph_id":1025,
        "properties":[{
            "property_id":2,
            "property_name":"id",
            "property_path":null,
            "unique":true
        }],
        "status":"Ready",
        "type_id":1,
        "unique":true,
        "version":0,
        "is_fulltext":false,
        "v_index_type":null,
        "is_primary_key":true
    }]}"
'''

'''
<<<<<<<< edge schema example >>>>>>>>
"person_knows_person":"{
    "graph_id":1025,
    "edge_type_id":24,
    "edge_type_name":"knows",
    "schema_version":0,
    "data_version":0,
    "status":"Normal",
    "properties":[{
        "property_id":51,
        "property_name":"creationDate",
        "property_type":"string",
        "type_length":0,
        "nullable":true,
        "primary_key":false,
        "default_value":{"object":null},
        "index_num":1,
        "order":0,
        "comment":"",
        "is_unique":false,
        "is_full_text":false,
        "v_index_type":null,
        "is_oid":false
    }],
    "vertex_type_pair":[{
        "from_id":1,
        "to_id":1,
        "status":"Normal",
        "from_name":"person",
        "to_name":"person"
    }],
    "partition_schema":{
        "method":"hash",
        "column":"_TYPE"
    },
    "rankable":false,
    "comment":"",
    "indexes":[{
        "index_id":25,
        "index_name":"knows_date_index",
        "index_type":"Edge",
        "graph_id":1025,
        "properties":[{
            "property_id":51,
            "property_name":"creationDate",
            "property_path":null,
            "unique":false
        }],
        "status":"Ready",
        "type_id":24,
        "unique":false,
        "version":0,
        "is_fulltext":false,
        "v_index_type":null,
        "is_primary_key":false
    }]
}"
'''