#  Copyright 2023 Fabarta Authors.
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import csv
import time

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import col, lit, row_number, split, concat_ws, to_timestamp, unix_millis
from pyspark.sql.types import ArrayType, StructType, ShortType
from pyspark.sql.window import Window

from config import *
from schema import *

INDEX_DIR = 'indexes'
TMP_RANK_COLUMN = '_tmp_rank'
EDGE_KEY_DIR = 'edge_key'
OUT_EDGE_KEY_DIR = 'out_edge_key'
IN_EDGE_KEY_DIR = 'in_edge_key'


class SparkProcessor:
    def __init__(self, config):
        self.config: Config = config
        self.spark = SparkSession.builder.master('local[{}]'.format(self.config.pool_size)) \
            .config('spark.driver.memory', self.config.driver_memory) \
            .config('spark.sql.shuffle.partitions', self.config.pool_size) \
            .config('spark.local.dir', self.config.spark_dir) \
            .appName('arc_direct_load').getOrCreate()
        self.spark.sparkContext.setLogLevel('ERROR')
        self.pk_oid_dfs = dict()  # vertex_name -> vertex pk to oid mapping dataframe

    ### sort original vertex csv
    def process_vertex_file(self, file_name: str, schema: dict):
        path = os.path.join(self.config.client_data_dir, file_name)
        if os.path.isfile(path):
            vertex_name: str = schema['vertex_type_name']
            start = time.time()
            columns = []
            if self.config.csv_has_headers:
                columns = self.get_csv_headers(path)
            fields = StructType(build_vertex_fields(schema, columns))
            df: DataFrame = self.spark.read.option('header', self.config.csv_has_headers) \
                  .option('delimiter', self.config.csv_delimiter) \
                  .option('quote', self.config.csv_quote) \
                  .option('escape', self.config.csv_escape) \
                  .schema(fields) \
                  .csv(path)
            pk_name = 'id'
            for prop in schema['properties']:
                if prop['primary_key']:
                    pk_name = prop['property_name']
                    break

            # check duplicate pk
            if self.config.check_duplicate:
                pk_window = Window.partitionBy(pk_name).orderBy(lit(1))
                df_group = df.withColumn('rn', row_number().over(pk_window))
                df_distinct = df_group.filter(col('rn') == 1).drop('rn')
                df_duplicates = df_group.filter(col('rn') > 1).drop('rn')
                if not df_duplicates.isEmpty():
                    df_duplicates.write.mode('overwrite') \
                        .option('header', True) \
                        .option('delimiter', self.config.csv_delimiter) \
                        .option('quote', self.config.csv_quote) \
                        .option('escape', self.config.csv_escape) \
                        .csv(os.path.join(self.config.error_dir, 'duplicate', vertex_name))
                    if self.config.on_duplicate == 'error':
                        raise Exception('duplicate vertex detected, vertex type is \"{}\", ' \
                                        'duplicate rows are recorded at \"{}\"'
                                        .format(vertex_name, self.config.error_dir))
                df = df_distinct

            # optimization: assign non-isolated(hotspot) vertexes with larger oid
            hotspot_file = os.path.join(self.config.hotspot_dir, file_name)
            if os.path.isfile(hotspot_file):
                df_hot: DataFrame = self.spark.read.option('header', self.config.csv_has_headers) \
                  .option('delimiter', self.config.csv_delimiter) \
                  .option('quote', self.config.csv_quote) \
                  .option('escape', self.config.csv_escape) \
                  .csv(hotspot_file)
                df_hot = df_hot.withColumn(TMP_RANK_COLUMN, lit(1))
                # sort by this temporary column in ascending order so that non-isolated vertexes
                # are at the end of the dataframe, thus they have larger oid
                df = df.join(df_hot, on=pk_name, how='left').sort(TMP_RANK_COLUMN, ascending=True)

            # assign monotonically increasing oid(row number)
            w = Window.partitionBy(lit(1)).orderBy(lit(1))
            df = df.withColumn(OID_COLUMN_NAME, row_number().over(w)).drop(TMP_RANK_COLUMN) 
            # sort df by oid in descending order
            df_sorted_by_oid = df.sort(OID_COLUMN_NAME, ascending=False)

            # extract mapping from pk to oid, which needs to be persisted
            df_sorted_by_oid.persist()
            df_pk2oid = df_sorted_by_oid.select([pk_name, OID_COLUMN_NAME])
            # store mapping from pk to oid for later usage: edge encoding needs this data
            self.pk_oid_dfs[vertex_name] = df_pk2oid

            # write out
            if not df_sorted_by_oid.isEmpty():
                df_sorted_by_oid.write.mode('overwrite') \
                    .option('header', True) \
                    .option('delimiter', self.config.csv_delimiter) \
                    .option('quote', self.config.csv_quote) \
                    .option('escape', self.config.csv_escape) \
                    .csv(os.path.join(self.config.sorted_csv_dir, vertex_name))

            end = time.time()
            print('sorting vertex file {} costs {:.2f} seconds'.format(path, end - start))

    ### sort original edge csv
    def process_edge_file(self, file_name: str, edge_schema: dict, edge_full_name: str, vertex_schemas: dict):
        path = os.path.join(self.config.client_data_dir, file_name)
        if os.path.isfile(path):
            start = time.time()
            src_vertex_name = edge_full_name.split('_')[0]
            dst_vertex_name = edge_full_name.split('_')[2]

            columns = []
            if self.config.csv_has_headers:
                columns = self.get_csv_headers(path)
            fields = StructType(build_edge_fields(edge_schema, columns, vertex_schemas, src_vertex_name, dst_vertex_name))
            df: DataFrame = self.spark.read.option('header', self.config.csv_has_headers) \
                  .option('delimiter', self.config.csv_delimiter) \
                  .option('quote', self.config.csv_quote) \
                  .option('escape', self.config.csv_escape) \
                  .schema(fields) \
                  .csv(path)
            columns = df.columns

            # check duplicate (src_id, dst_id[, rank])
            if self.config.check_duplicate:
                if edge_schema['rankable']:
                    group_window = Window.partitionBy([EDGE_SRC_ID_COLUMN, EDGE_DST_ID_COLUMN, EDGE_RANK_COLUMN]).orderBy(lit(1))
                else:
                   group_window = Window.partitionBy([EDGE_SRC_ID_COLUMN, EDGE_DST_ID_COLUMN]).orderBy(lit(1)) 
                df_group = df.withColumn('rn', row_number().over(group_window))
                df_distinct = df_group.filter(col('rn') == 1).drop('rn')
                df_duplicates = df_group.filter(col('rn') > 1).drop('rn')
                if not df_duplicates.isEmpty():
                    df_duplicates.write.mode('overwrite') \
                        .option('header', True) \
                        .option('delimiter', self.config.csv_delimiter) \
                        .option('quote', self.config.csv_quote) \
                        .option('escape', self.config.csv_escape) \
                        .csv(os.path.join(self.config.error_dir, 'duplicate', edge_full_name))
                    if self.config.on_duplicate == 'error':
                        raise Exception('duplicate edge detected, edge type is \"{}\", '\
                                        'duplicate rows are recorded at \"{}\"'
                                        .format(edge_full_name, self.config.error_dir))
                df = df_distinct

            # join with vertex id
            src_df_pk2oid = self.pk_oid_dfs[src_vertex_name].withColumnRenamed('id', EDGE_SRC_ID_COLUMN)
            dst_df_pk2oid = self.pk_oid_dfs[dst_vertex_name].withColumnRenamed('id', EDGE_DST_ID_COLUMN)

            # a left join returns all values from the left relation and the matched values from the right relation,
            # or appends NULL if there is no match
            # after join, df columns will be like (src_id, dst_id, properties[, rank], src_oid, dst_oid)
            df = df \
                .join(src_df_pk2oid, on=EDGE_SRC_ID_COLUMN, how='left') \
                .withColumnRenamed(OID_COLUMN_NAME, 'src_oid') \
                .join(dst_df_pk2oid, on=EDGE_DST_ID_COLUMN, how='left') \
                .withColumnRenamed(OID_COLUMN_NAME, 'dst_oid')

            # record dangling edges
            df_null = df.filter(col('src_oid').isNull() | col('dst_oid').isNull())
            if not df_null.isEmpty():
                df_null.select(columns).write.mode('overwrite') \
                    .option('header', True) \
                    .option('delimiter', self.config.csv_delimiter) \
                    .option('quote', self.config.csv_quote) \
                    .option('escape', self.config.csv_escape) \
                    .csv(os.path.join(self.config.error_dir, 'dangling', edge_full_name))
                print('dangling edges detected, edge type is \"{}\"; dangling edges are recorded at \"{}\"'.format(edge_full_name, self.config.error_dir))
                # drop dangling edges
                df = df.filter(col('src_oid').isNotNull() & col('dst_oid').isNotNull())

            df = df \
                .drop(EDGE_SRC_ID_COLUMN) \
                .drop(EDGE_DST_ID_COLUMN) \
                .withColumnRenamed('src_oid', EDGE_SRC_ID_COLUMN) \
                .withColumnRenamed('dst_oid', EDGE_DST_ID_COLUMN) \
                .select(columns)

            # convert rank from string to unix timestamp in MILLISECONDS if format is given
            if edge_schema['rankable'] and self.config.is_timestamp_rank(edge_schema['edge_type_name']):
                ts_fmt = self.config.get_timestamp_rank_fmt(edge_schema['edge_type_name'])
                df = df.withColumn('unix_timestamp', unix_millis(to_timestamp(EDGE_RANK_COLUMN, ts_fmt))) \
                    .drop(EDGE_RANK_COLUMN).withColumnRenamed('unix_timestamp', EDGE_RANK_COLUMN)

            if not df.isEmpty():
                # sort by (src_id, dst_id[, rank]) in descending order because oid and rank will be encoded with its complement
                # which is sorted in ascending order
                if edge_schema['rankable']:
                    df_sorted_out_edge = df.sort([EDGE_SRC_ID_COLUMN, EDGE_DST_ID_COLUMN, EDGE_RANK_COLUMN],
                                            ascending=[False, False, False])
                else:
                    df_sorted_out_edge = df.sort([EDGE_SRC_ID_COLUMN, EDGE_DST_ID_COLUMN],
                                            ascending=[False, False])
                    
                # write out
                df_sorted_out_edge.write.mode('overwrite') \
                    .option('header', True) \
                    .option('delimiter', self.config.csv_delimiter) \
                    .option('quote', self.config.csv_quote) \
                    .option('escape', self.config.csv_escape) \
                    .csv(os.path.join(self.config.sorted_csv_dir, edge_full_name, 'out'))

                if self.config.reverse_edge:
                    # sort by (dst_id, src_id[, rank]) in descending order because oid and rank will be encoded with its complement
                    # which is sorted in ascending order
                    if edge_schema['rankable']:
                        df_sorted_in_edge = df.sort([EDGE_DST_ID_COLUMN, EDGE_SRC_ID_COLUMN, EDGE_RANK_COLUMN],
                                                ascending=[False, False, False])
                    else:
                        df_sorted_in_edge = df.sort([EDGE_DST_ID_COLUMN, EDGE_SRC_ID_COLUMN],
                                                ascending=[False, False])
                    # write out
                    df_sorted_in_edge.write.mode('overwrite') \
                        .option('header', True) \
                        .option('delimiter', self.config.csv_delimiter) \
                        .option('quote', self.config.csv_quote) \
                        .option('escape', self.config.csv_escape) \
                        .csv(os.path.join(self.config.sorted_csv_dir, edge_full_name, 'in'))

            end = time.time()
            print('sorting edge file {} costs {:.2f} seconds'.format(path, end - start))

    ### sort encoded pk-oid index
    def process_pkoid_dir(self, path, index_name):
        if os.path.isdir(path):
            start = time.time()
            schema = StructType([StructField('pk_str', StringType(), False), StructField('oid', StringType(), False)])
            df: DataFrame = self.spark.read.option('header', False) \
                  .option('delimiter', self.config.csv_delimiter) \
                  .option('quote', self.config.csv_quote) \
                  .option('escape', self.config.csv_escape) \
                  .schema(schema) \
                  .csv(path)
            
            # convert encoded key from string to array of short type integer
            df_split = df \
                .withColumn('pk_str_list', split(df['pk_str'], ',')) \
                .drop('pk_str') \
                .withColumn('pk_short_list', col('pk_str_list').cast(ArrayType(ShortType()))) \
                .drop('pk_str_list')
            
            df_sorted = df_split.sort('pk_short_list')
            df_sorted = df_sorted.withColumn('pk', concat_ws(',', col('pk_short_list'))).drop('pk_short_list').select(['pk', 'oid'])
            df_sorted.write.mode('overwrite') \
                .option('header', True) \
                .option('delimiter', self.config.csv_delimiter) \
                .option('quote', self.config.csv_quote) \
                .option('escape', self.config.csv_escape) \
                .csv(os.path.join(self.config.sorted_csv_dir, INDEX_DIR, index_name))

            end = time.time()
            print('sorting index directory {} costs {:.2f} seconds'.format(path, end - start))
            
    ### sort encoded index directory
    def process_normal_index_dir(self, path, index_name):
        if os.path.isdir(path):
            start = time.time()
            schema = StructType([StructField('index', StringType(), False)])
            df: DataFrame = self.spark.read.option('header', False) \
                  .option('delimiter', self.config.csv_delimiter) \
                  .option('quote', self.config.csv_quote) \
                  .option('escape', self.config.csv_escape) \
                  .schema(schema) \
                  .csv(path)
            
            # convert encoded key from string to array of short type integer
            df_split = df \
                .withColumn('index_str_list', split(df['index'], ',')) \
                .drop('index') \
                .withColumn('index_short_list', col('index_str_list').cast(ArrayType(ShortType()))) \
                .drop('index_str_list')
            
            df_sorted = df_split.sort('index_short_list')
            df_sorted = df_sorted.withColumn('index', concat_ws(',', col('index_short_list'))).drop('index_short_list')
            df_sorted.write.mode('overwrite') \
                .option('header', True) \
                .option('delimiter', self.config.csv_delimiter) \
                .option('quote', self.config.csv_quote) \
                .option('escape', self.config.csv_escape) \
                .csv(os.path.join(self.config.sorted_csv_dir, INDEX_DIR, index_name))

            end = time.time()
            print('sorting index directory {} costs {:.2f} seconds'.format(path, end - start))

    ### sort encoded edge key directory
    def process_edge_key_dir(self, path, start_vertex, direction):
        if os.path.isdir(path):
            start = time.time()
            schema = StructType([StructField('edge_key', StringType(), False), 
                                 StructField('value', StringType(), False)])
            df: DataFrame = self.spark.read.option('header', False) \
                  .option('delimiter', self.config.csv_delimiter) \
                  .option('quote', self.config.csv_quote) \
                  .option('escape', self.config.csv_escape) \
                  .schema(schema) \
                  .csv(path)
            
            # convert encoded key from string to array of short type integer
            df_split = df \
                .withColumn('edge_key_str_list', split(df['edge_key'], ',')) \
                .drop('edge_key') \
                .withColumn('edge_key_short_list', col('edge_key_str_list').cast(ArrayType(ShortType()))) \
                .drop('edge_key_str_list')
            
            df_sorted = df_split.sort('edge_key_short_list')
            df_sorted = df_sorted.withColumn('edge_key', concat_ws(',', col('edge_key_short_list'))).drop('edge_key_short_list')
            df_sorted = df_sorted.select(['edge_key', 'value'])
            df_sorted.write.mode('overwrite') \
                .option('header', True) \
                .option('delimiter', self.config.csv_delimiter) \
                .option('quote', self.config.csv_quote) \
                .option('escape', self.config.csv_escape) \
                .csv(os.path.join(self.config.sorted_csv_dir, EDGE_KEY_DIR, direction, start_vertex))

            end = time.time()
            print('sorting edge key directory {} costs {:.2f} seconds'.format(path, end - start)) 
        
    def get_csv_headers(self, path):
        with open(path, 'r') as f:
            reader = csv.reader(f, delimiter=self.config.csv_delimiter, quotechar=self.config.csv_quote,
                                escapechar=self.config.csv_escape, quoting=0)
            for row in reader:
                return row
